"""Whether to pull the `generate` method out from the `DecoderGPT` class
and make it a standalone function/class is a design issue. Putting it under
the `DecoderGPT` class makes it easier to access the `self` attributes."""
